//
//  Copyright (C) 2001,2002 HorizonLive.com, Inc.  All Rights Reserved.
//  Copyright (C) 2001,2002 Constantin Kaplinsky.  All Rights Reserved.
//  Copyright (C) 2000 Tridia Corporation.  All Rights Reserved.
//  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
//
//  This is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This software is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this software; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
//  USA.
//
package com.sshtools.sshvnc;

import java.awt.AWTEvent;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.image.ColorModel;
import java.awt.image.DirectColorModel;
import java.awt.image.MemoryImageSource;
import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.util.zip.Inflater;

import javax.swing.JComponent;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.sshtools.common.ui.UIUtil;

//
// VncCanvas is a subclass of Canvas which draws a VNC desktop on it.
//

public class VncCanvas
    extends JComponent
    implements KeyListener, MouseListener, MouseMotionListener {
    /**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	private static Log log = LogFactory.getLog(VncCanvas.class);

//  VncViewer viewer;
  SshVNCViewer viewer = null;
  RfbProto rfb = null;
  ColorModel cm8, cm24;
  Color[] colors  = null;
  int bytesPixel;

  Image memImage  = null;
  Graphics memGraphics  = null;

  Image rawPixelsImage  = null;
  MemoryImageSource pixelsSource  = null;
  byte[] pixels8;
  int[] pixels24;

  // Zlib encoder's data.
  byte[] zlibBuf;
  int zlibBufLen = 0;
  Inflater zlibInflater  = null;

  // Tight encoder's data.
  final static int tightZlibBufferSize = 512;
  Inflater[] tightInflaters  = null;
  
  

  // Since JPEG images are loaded asynchronously, we have to remember
  // their position in the framebuffer. Also, this jpegRect object is
  // used for synchronization between the rfbThread and a JVM's thread
  // which decodes and loads JPEG images.
  Rectangle jpegRect = null;

  // True if we process keyboard and mouse events.
  boolean inputEnabled;

  private final static long VNC_EVENTS = AWTEvent.KEY_EVENT_MASK
      | AWTEvent.FOCUS_EVENT_MASK | AWTEvent.ACTION_EVENT_MASK
      | AWTEvent.MOUSE_MOTION_EVENT_MASK | AWTEvent.MOUSE_EVENT_MASK;

  //
  // The constructor.
  //

  public VncCanvas(SshVNCViewer v) throws IOException {

    enableEvents(VNC_EVENTS);
    viewer = v;
    rfb = viewer.getRFB();

    tightInflaters = new Inflater[4];

    cm8 = new DirectColorModel(8, 7, (7 << 3), (3 << 6));
    cm24 = new DirectColorModel(24, 0xFF0000, 0x00FF00, 0x0000FF);

    colors = new Color[256];
    for (int i = 0; i < 256; i++) {
      colors[i] = new Color(cm8.getRGB(i));

    }
    setPixelFormat();

    inputEnabled = false;
    if (!viewer.getOptions().isViewOnly()) {
      enableInput(true);

      // Keyboard listener is enabled even in view-only mode, to catch
      // 'r' or 'R' key presses used to request screen update.
    }
    addKeyListener(this);

    /*
     *  we have to make sure the tab key stays within the component
     *  
     *  04/12/06 Bob Baske
     *  Modified to handle java versions greater then 1.3. Previously 
     *  the code was written to handle just 1.4.
     */

    String version = System.getProperty("java.version").substring(0, 3);
    float jdk = new Float(version).floatValue();

    log.debug("VncCanvas" + ":" + "VncCanvas"
			+ " version: " + version + " jdk: " + jdk);
    if (jdk > 1.3) { 
      try {
        Class[] params = new Class[] {
            boolean.class};
        VncCanvas.class.getMethod("setFocusable", params).invoke(this,
            new Object[] {new Boolean(true)});
        VncCanvas.class.getMethod("setFocusTraversalKeysEnabled",
                                  params).invoke(this,
                                                 new Object[] {new Boolean(false)});
      }
      catch (Exception e) {
        System.err.println(
            "VncCanvas: unable to reset focus handling for java version "
            + version);
        e.printStackTrace();
      }
    }

  }

  //
  // Callback methods to determine geometry of our Component.
  //

  public Dimension getPreferredSize() {
    return new Dimension(rfb.framebufferWidth, rfb.framebufferHeight);
  }

  public Dimension getMinimumSize() {
    return new Dimension(rfb.framebufferWidth, rfb.framebufferHeight);
  }

  public Dimension getMaximumSize() {
    return new Dimension(rfb.framebufferWidth, rfb.framebufferHeight);
  }

  //
  // All painting is performed here.
  //

//  public void update(Graphics g) {
//    paint(g);
//  }
  
  private Rectangle r = null;

  public void paintComponent(Graphics g) {
    synchronized (memImage) {
      g.drawImage(memImage, 0, 0, null);
    }
    if (showSoftCursor) {
      int x0 = cursorX - hotX, y0 = cursorY - hotY;
      r = new Rectangle(x0, y0, cursorWidth, cursorHeight);
      if (r.intersects(g.getClipBounds())) {
        g.drawImage(softCursor, x0, y0, null);
      }
    }
  }

  //
  // Override the ImageObserver interface method to handle drawing of
  // JPEG-encoded data.
  //

  public boolean imageUpdate(Image img, int infoflags,
                             int x, int y, int width, int height) {
    if ( (infoflags & (ALLBITS | ABORT)) == 0) {
      return true; // We need more image data.
    }
    else {
      // If the whole image is available, draw it now.
      if ( (infoflags & ALLBITS) != 0) {
        if (jpegRect != null) {
          synchronized (jpegRect) {
            memGraphics.drawImage(img, jpegRect.x, jpegRect.y, null);
            scheduleRepaint(jpegRect.x, jpegRect.y,
                            jpegRect.width, jpegRect.height);
            jpegRect.notify();
          }
        }
      }
      return false; // All image data was processed.
    }
  }

  //
  // Start/stop receiving mouse events. Keyboard events are received
  // even in view-only mode, because we want to map the 'r' key to the
  // screen refreshing function.
  //

  public synchronized void enableInput(boolean enable) {
    if (enable && !inputEnabled) {
      inputEnabled = true;
      addMouseListener(this);
      addMouseMotionListener(this);
    }
    else if (!enable && inputEnabled) {
      inputEnabled = false;
      removeMouseListener(this);
      removeMouseMotionListener(this);
    }
  }

  public void setPixelFormat() throws IOException {
      if (log.isDebugEnabled()) {
        log.debug("enter :" + "VncCanvas" + ":" + "setPixelFormat");
    }
      log.debug("VncCanvas" + ":" + "setPixelFormat"
              + " isEightBitColors: " + viewer.getOptions().isEightBitColors());
    if (viewer.getOptions().isEightBitColors()) {
      rfb.writeSetPixelFormat(8, 8, false, true, 7, 7, 3, 0, 3, 6);
      bytesPixel = 1;
    }
    else {
      rfb.writeSetPixelFormat(32, 24, false, true, 255, 255, 255, 16, 8, 0);
      bytesPixel = 4;
    }
    updateFramebufferSize();
  }

  void updateFramebufferSize() {
    // Useful shortcuts.
    int fbWidth = rfb.framebufferWidth;
    int fbHeight = rfb.framebufferHeight;

    // Create new off-screen image either if it does not exist, or if
    // its geometry should be changed. It's not necessary to replace
    // existing image if only pixel format should be changed.
    if (memImage == null) {
      memImage = viewer.createImage(fbWidth, fbHeight);
      memGraphics = memImage.getGraphics();
    }
    else if (memImage.getWidth(null) != fbWidth ||
             memImage.getHeight(null) != fbHeight) {
      synchronized (memImage) {
        memImage = viewer.createImage(fbWidth, fbHeight);
        memGraphics = memImage.getGraphics();
      }
    }

    // Images with raw pixels should be re-allocated on every change
    // of geometry or pixel format.
    if (bytesPixel == 1) {
      pixels24 = null;
      pixels8 = new byte[fbWidth * fbHeight];

      pixelsSource =
          new MemoryImageSource(fbWidth, fbHeight, cm8, pixels8, 0, fbWidth);
    }
    else {
      pixels8 = null;
      pixels24 = new int[fbWidth * fbHeight];

      pixelsSource =
          new MemoryImageSource(fbWidth, fbHeight, cm24, pixels24, 0, fbWidth);
    }
    pixelsSource.setAnimated(true);
    rawPixelsImage = createImage(pixelsSource);

    // Update the size of desktop containers.
    if (viewer.getScroller() != null) {
      resizeDesktopFrame();
    }
    else {
      setSize(fbWidth, fbHeight);
    }
    viewer.moveFocusToDesktop();
  }

  void resizeDesktopFrame() {
    // If we are not setting the screen size explicitly then resize
    if (viewer.getScreenSizePolicy() ==
        SshVNCPanel.PROFILE_SCREEN_REMOTE_DESKTOP) {

//      System.out.print("*** Resizing to " + rfb.framebufferWidth + "," +
//                       rfb.framebufferHeight);

      setSize(rfb.framebufferWidth, rfb.framebufferHeight);

      if (viewer.getFrame() != null && viewer.isFrameResizeable()) {
        viewer.getFrame().pack();
      }
      else {
        revalidate();
      }
    }
    /* else {
       setSize(viewer.getScreenSize());
       viewer.getFrame().pack();
       viewer.getFrame().setSize(viewer.getScreenSize());
       repaint();
       //viewer.getFrame().invalidate();
       //viewer.revalidate();
       //revalidate();
     }*/

    // Recenter the viewer upon resize completion
    UIUtil.positionComponent(UIUtil.CENTER, viewer);
  }

  //
  // processNormalProtocol() - executed by the rfbThread to deal with the
  // RFB socket.
  //

  public void processNormalProtocol() throws Exception {
    if (log.isDebugEnabled()) {
        log.debug("enter :" + "VncCanvas" + ":" + "processNormalProtocol");
    }
    
    // Start/stop session recording if necessary.
    viewer.checkRecordingStatus();

    log.debug("VncCanvas" + ":" + "processNormalProtocol"
            + " ---- call writeFramebufferUpdateRequest at top of processNormalProtocol ");
    rfb.writeFramebufferUpdateRequest(0, 0, rfb.framebufferWidth,
                                      rfb.framebufferHeight, false);

    //
    // main dispatch loop
    //

    while (true) {
      // Read message type from the server.
      int msgType = rfb.readServerMessageType();
      
      log.debug("VncCanvas" + ":" + "processNormalProtocol"
			+ " message type:" + msgType);

      // Process the message depending on its type.
      switch (msgType) {
        case RfbProto.FramebufferUpdate:
          rfb.readFramebufferUpdate();

          boolean cursorPosReceived = false;

          for (int i = 0; i < rfb.updateNRects; i++) {                                            
            rfb.readFramebufferUpdateRectHdr();
            int rx = rfb.updateRectX, ry = rfb.updateRectY;
            int rw = rfb.updateRectW, rh = rfb.updateRectH;

            if (rfb.updateRectEncoding == RfbProto.EncodingLastRect) {
                log.debug("VncCanvas" + ":" + "processNormalProtocol"
                        + " " + "rfb.updateRectEncoding == rfb.EncodingLastRect");
              break;
            }

            if (rfb.updateRectEncoding == RfbProto.EncodingNewFBSize) {
                log.debug("VncCanvas" + ":" + "processNormalProtocol"
                        + " " + "rfb.updateRectEncoding == rfb.EncodingNewFBSize");
              rfb.setFramebufferSize(rw, rh);                                                              
              updateFramebufferSize();
              break;
            }
            /*
             * EncodingXCursor = 0xFFFFFF10,
             * EncodingRichCursor = 0xFFFFFF11 (-239),
             */
            if (rfb.updateRectEncoding == RfbProto.EncodingXCursor ||                                  
                rfb.updateRectEncoding == RfbProto.EncodingRichCursor) {
                log.debug("VncCanvas" + ":" + "processNormalProtocol"
                        + " " + "rfb.updateRectEncoding == rfb.EncodingXCursor || rfb.updateRectEncoding == rfb.EncodingRichCursor");
              handleCursorShapeUpdate(rfb.updateRectEncoding, rx, ry, rw, rh);
              continue;
            }

            if (rfb.updateRectEncoding == RfbProto.EncodingPointerPos) {
                log.debug("VncCanvas" + ":" + "processNormalProtocol"
                        + " " + "rfb.updateRectEncoding == rfb.EncodingPointerPos");
              softCursorMove(rx, ry);
              cursorPosReceived = true; 
              continue;
            }

            switch (rfb.updateRectEncoding) {
              case RfbProto.EncodingRaw:
                handleRawRect(rx, ry, rw, rh);
                break;
              case RfbProto.EncodingCopyRect:
                handleCopyRect(rx, ry, rw, rh);
                break;
              case RfbProto.EncodingRRE:
                handleRRERect(rx, ry, rw, rh);
                break;
              case RfbProto.EncodingCoRRE:
                handleCoRRERect(rx, ry, rw, rh);
                break;
              case RfbProto.EncodingHextile:
                handleHextileRect(rx, ry, rw, rh);
                break;
              case RfbProto.EncodingZlib:
                handleZlibRect(rx, ry, rw, rh);
                break;
              case RfbProto.EncodingTight:
                handleTightRect(rx, ry, rw, rh);
                break;
              default:
                log.error("VncCanvas" + ":" + "processNormalProtocol"
                        + " " + "Unknown RFB rectangle encoding " +
                                    rfb.updateRectEncoding);
                throw new Exception("Unknown RFB rectangle encoding " +
                                    rfb.updateRectEncoding);
            }
          }
          
          boolean fullUpdateNeeded = false;

          // Start/stop session recording if necessary. Request full
          // update if a new session file was opened.
          if (viewer.checkRecordingStatus()) {
            fullUpdateNeeded = true;
          }

          // Defer framebuffer update request if necessary. But wake up
          // immediately on keyboard or mouse event. Also, don't sleep
          // if there is some data to receive, or if the last update
          // included a PointerPos message.     
          if (viewer.getDeferUpdateRequests() > 0 &&
              rfb.is.available() == 0 && !cursorPosReceived) {
            synchronized (rfb) {
              try {
                rfb.wait(viewer.getDeferUpdateRequests());
              }
              catch (InterruptedException e) {
              }
            }
          }
                  
          // Before requesting framebuffer update, check if the pixel
          // format should be changed. If it should, request full update
          // instead of an incremental one.
          if (viewer.getOptions().isEightBitColors() != (bytesPixel == 1)) {
            setPixelFormat();
            fullUpdateNeeded = true;
          }
          
          rfb.writeFramebufferUpdateRequest(0, 0, rfb.framebufferWidth,
                                            rfb.framebufferHeight,
                                            !fullUpdateNeeded);

          break;

        case RfbProto.SetColourMapEntries:
          throw new Exception("Can't handle SetColourMapEntries message");

        case RfbProto.Bell:
          Toolkit.getDefaultToolkit().beep();
          break;
          
        case RfbProto.ServerCutText:
        	String s = rfb.readServerCutText();
        	viewer.getStatusBar().setClientClipboard(s);
          break;
        /*
         * 3/14/11 Bob Baske 
         *
         * Custom message - consists of message id (8) and nothing else. Will be immediately
         * followed by a Sync message (9) which has the package timestamp.
         */
        case RfbProto.SyncMessage:
            break;  
        
        case RfbProto.Sync:
          long ms = rfb.readServerSyncTimestamp();
          if (ms == -1) {
        	  viewer.getStatusBar().setPlaybackControlsActive(true);
//        	  viewer.selectEvent(ms);
          } else {
        	  viewer.getStatusBar().setNow(ms);
//        	  viewer.selectEvent(ms);
          }
          break;
          
        default:
          log.debug("VncCanvas" + ":" + "processNormalProtocol"
                + " Unknown RFB message type: " + msgType);
          throw new Exception("Unknown RFB message type " + msgType);
      }
    }
  }

  //
  // Handle a raw rectangle. The second form with paint==false is used
  // by the Hextile decoder for raw-encoded tiles.
  //

  void handleRawRect(int x, int y, int w, int h) throws IOException {
    if (log.isDebugEnabled()) {
        log.debug("enter :" + "VncCanvas" + ":" + "handleRawRect");
    }
    handleRawRect(x, y, w, h, true);
  }

  void handleRawRect(int x, int y, int w, int h, boolean paint) throws
      IOException {

    if (bytesPixel == 1) {
      for (int dy = y; dy < y + h; dy++) {
        rfb.is.readFully(pixels8, dy * rfb.framebufferWidth + x, w);
        if (rfb.rec != null) {
          rfb.rec.write(pixels8, dy * rfb.framebufferWidth + x, w);
        }
      }
    }
    else {
      byte[] buf = new byte[w * 4];
      int i, offset;
      for (int dy = y; dy < y + h; dy++) {
        rfb.is.readFully(buf);
        if (rfb.rec != null) {
          rfb.rec.write(buf);
        }
        offset = dy * rfb.framebufferWidth + x;
        for (i = 0; i < w; i++) {
          pixels24[offset + i] =
              (buf[i * 4 + 2] & 0xFF) << 16 |
              (buf[i * 4 + 1] & 0xFF) << 8 |
              (buf[i * 4] & 0xFF);
        }
      }
      buf = null;
    }

    handleUpdatedPixels(x, y, w, h);
    if (paint) {
      scheduleRepaint(x, y, w, h);
    }
  }

  //
  // Handle a CopyRect rectangle.
  //

  void handleCopyRect(int x, int y, int w, int h) throws IOException {

    rfb.readCopyRect();
    memGraphics.copyArea(rfb.copyRectSrcX, rfb.copyRectSrcY, w, h,
                         x - rfb.copyRectSrcX, y - rfb.copyRectSrcY);

    scheduleRepaint(x, y, w, h);
  }

  //
  // Handle an RRE-encoded rectangle.
  //

  void handleRRERect(int x, int y, int w, int h) throws IOException {

    int nSubrects = rfb.is.readInt();

    byte[] bg_buf = new byte[bytesPixel];
    rfb.is.readFully(bg_buf);
    Color pixel;
    if (bytesPixel == 1) {
      pixel = colors[bg_buf[0] & 0xFF];
    }
    else {
      pixel = new Color(bg_buf[2] & 0xFF, bg_buf[1] & 0xFF, bg_buf[0] & 0xFF);
    }
    memGraphics.setColor(pixel);
    memGraphics.fillRect(x, y, w, h);

    byte[] buf = new byte[nSubrects * (bytesPixel + 8)];
    rfb.is.readFully(buf);
    DataInputStream ds = new DataInputStream(new ByteArrayInputStream(buf));

    if (rfb.rec != null) {
      rfb.rec.writeIntBE(nSubrects);
      rfb.rec.write(bg_buf);
      rfb.rec.write(buf);
    }

    int sx, sy, sw, sh;

    for (int j = 0; j < nSubrects; j++) {
      if (bytesPixel == 1) {
        pixel = colors[ds.readUnsignedByte()];
      }
      else {
        ds.skip(4);
        pixel = new Color(buf[j * 12 + 2] & 0xFF,
                          buf[j * 12 + 1] & 0xFF,
                          buf[j * 12] & 0xFF);
      }
      sx = x + ds.readUnsignedShort();
      sy = y + ds.readUnsignedShort();
      sw = ds.readUnsignedShort();
      sh = ds.readUnsignedShort();

      memGraphics.setColor(pixel);
      memGraphics.fillRect(sx, sy, sw, sh);
    }

    scheduleRepaint(x, y, w, h);
    bg_buf = null;
    buf = null;
    pixel = null;
  }

  //
  // Handle a CoRRE-encoded rectangle.
  //

  void handleCoRRERect(int x, int y, int w, int h) throws IOException {
    int nSubrects = rfb.is.readInt();

    byte[] bg_buf = new byte[bytesPixel];
    rfb.is.readFully(bg_buf);
    Color pixel;
    if (bytesPixel == 1) {
      pixel = colors[bg_buf[0] & 0xFF];
    }
    else {
      pixel = new Color(bg_buf[2] & 0xFF, bg_buf[1] & 0xFF, bg_buf[0] & 0xFF);
    }
    memGraphics.setColor(pixel);
    memGraphics.fillRect(x, y, w, h);

    byte[] buf = new byte[nSubrects * (bytesPixel + 4)];
    rfb.is.readFully(buf);

    if (rfb.rec != null) {
      rfb.rec.writeIntBE(nSubrects);
      rfb.rec.write(bg_buf);
      rfb.rec.write(buf);
    }

    int sx, sy, sw, sh;
    int i = 0;

    for (int j = 0; j < nSubrects; j++) {
      if (bytesPixel == 1) {
        pixel = colors[buf[i++] & 0xFF];
      }
      else {
        pixel = new Color(buf[i + 2] & 0xFF, buf[i + 1] & 0xFF, buf[i] & 0xFF);
        i += 4;
      }
      sx = x + (buf[i++] & 0xFF);
      sy = y + (buf[i++] & 0xFF);
      sw = buf[i++] & 0xFF;
      sh = buf[i++] & 0xFF;

      memGraphics.setColor(pixel);
      memGraphics.fillRect(sx, sy, sw, sh);
    }

    scheduleRepaint(x, y, w, h);
    bg_buf = null;
    buf = null;
    pixel = null;
  }

  //
  // Handle a Hextile-encoded rectangle.
  //

  // These colors should be kept between handleHextileSubrect() calls.
  private Color hextile_bg, hextile_fg;

  void handleHextileRect(int x, int y, int w, int h) throws IOException {

    hextile_bg = new Color(0);
    hextile_fg = new Color(0);

    for (int ty = y; ty < y + h; ty += 16) {
      int th = 16;
      if (y + h - ty < 16) {
        th = y + h - ty;

      }
      for (int tx = x; tx < x + w; tx += 16) {
        int tw = 16;
        if (x + w - tx < 16) {
          tw = x + w - tx;

        }
        handleHextileSubrect(tx, ty, tw, th);
      }

      // Finished with a row of tiles, now let's show it.
      scheduleRepaint(x, y, w, h);
    }
  }

  //
  // Handle one tile in the Hextile-encoded data.
  //

  void handleHextileSubrect(int tx, int ty, int tw, int th) throws IOException {
    int subencoding = rfb.is.readUnsignedByte();
    if (rfb.rec != null) {
      rfb.rec.writeByte(subencoding);
    }

    // Is it a raw-encoded sub-rectangle?
    if ( (subencoding & rfb.HextileRaw) != 0) {
      handleRawRect(tx, ty, tw, th, false);
      return;
    }

    // Read and draw the background if specified.
    byte[] cbuf = new byte[bytesPixel];
    if ( (subencoding & rfb.HextileBackgroundSpecified) != 0) {
      rfb.is.readFully(cbuf);
      if (bytesPixel == 1) {
        hextile_bg = colors[cbuf[0] & 0xFF];
      }
      else {
        hextile_bg = new Color(cbuf[2] & 0xFF, cbuf[1] & 0xFF, cbuf[0] & 0xFF);
      }
      if (rfb.rec != null) {
        rfb.rec.write(cbuf);
      }
    }
    memGraphics.setColor(hextile_bg);
    memGraphics.fillRect(tx, ty, tw, th);

    // Read the foreground color if specified.
    if ( (subencoding & rfb.HextileForegroundSpecified) != 0) {
      rfb.is.readFully(cbuf);
      if (bytesPixel == 1) {
        hextile_fg = colors[cbuf[0] & 0xFF];
      }
      else {
        hextile_fg = new Color(cbuf[2] & 0xFF, cbuf[1] & 0xFF, cbuf[0] & 0xFF);
      }
      if (rfb.rec != null) {
        rfb.rec.write(cbuf);
      }
    }

    // Done with this tile if there is no sub-rectangles.
    if ( (subencoding & rfb.HextileAnySubrects) == 0) {
      return;
    }

    int nSubrects = rfb.is.readUnsignedByte();
    int bufsize = nSubrects * 2;
    if ( (subencoding & rfb.HextileSubrectsColoured) != 0) {
      bufsize += nSubrects * bytesPixel;
    }
    byte[] buf = new byte[bufsize];
    rfb.is.readFully(buf);
    if (rfb.rec != null) {
      rfb.rec.writeByte(nSubrects);
      rfb.rec.write(buf);
    }

    int b1, b2, sx, sy, sw, sh;
    int i = 0;

    if ( (subencoding & rfb.HextileSubrectsColoured) == 0) {

      // Sub-rectangles are all of the same color.
      memGraphics.setColor(hextile_fg);
      for (int j = 0; j < nSubrects; j++) {
        b1 = buf[i++] & 0xFF;
        b2 = buf[i++] & 0xFF;
        sx = tx + (b1 >> 4);
        sy = ty + (b1 & 0xf);
        sw = (b2 >> 4) + 1;
        sh = (b2 & 0xf) + 1;
        memGraphics.fillRect(sx, sy, sw, sh);
      }
    }
    else if (bytesPixel == 1) {

      // BGR233 (8-bit color) version for colored sub-rectangles.
      for (int j = 0; j < nSubrects; j++) {
        hextile_fg = colors[buf[i++] & 0xFF];
        b1 = buf[i++] & 0xFF;
        b2 = buf[i++] & 0xFF;
        sx = tx + (b1 >> 4);
        sy = ty + (b1 & 0xf);
        sw = (b2 >> 4) + 1;
        sh = (b2 & 0xf) + 1;
        memGraphics.setColor(hextile_fg);
        memGraphics.fillRect(sx, sy, sw, sh);
      }

    }
    else {

      // Full-color (24-bit) version for colored sub-rectangles.
      for (int j = 0; j < nSubrects; j++) {
        hextile_fg = new Color(buf[i + 2] & 0xFF,
                               buf[i + 1] & 0xFF,
                               buf[i] & 0xFF);
        i += 4;
        b1 = buf[i++] & 0xFF;
        b2 = buf[i++] & 0xFF;
        sx = tx + (b1 >> 4);
        sy = ty + (b1 & 0xf);
        sw = (b2 >> 4) + 1;
        sh = (b2 & 0xf) + 1;
        memGraphics.setColor(hextile_fg);
        memGraphics.fillRect(sx, sy, sw, sh);
      }
    }
    
    cbuf = null;
    buf = null;
    
  }

  //
  // Handle a Zlib-encoded rectangle.
  //

  void handleZlibRect(int x, int y, int w, int h) throws Exception {

    int nBytes = rfb.is.readInt();

    if (zlibBuf == null || zlibBufLen < nBytes) {
      zlibBufLen = nBytes * 2;
      zlibBuf = new byte[zlibBufLen];
    }

    rfb.is.readFully(zlibBuf, 0, nBytes);

    if (rfb.rec != null && rfb.recordFromBeginning) {
      rfb.rec.writeIntBE(nBytes);
      rfb.rec.write(zlibBuf, 0, nBytes);
    }

    if (zlibInflater == null) {
      zlibInflater = new Inflater();
    }
    zlibInflater.setInput(zlibBuf, 0, nBytes);

    if (bytesPixel == 1) {
      for (int dy = y; dy < y + h; dy++) {
        zlibInflater.inflate(pixels8, dy * rfb.framebufferWidth + x, w);
        if (rfb.rec != null && !rfb.recordFromBeginning) {
          rfb.rec.write(pixels8, dy * rfb.framebufferWidth + x, w);
        }
      }
    }
    else {
      byte[] buf = new byte[w * 4];
      int i, offset;
      for (int dy = y; dy < y + h; dy++) {
        zlibInflater.inflate(buf);
        offset = dy * rfb.framebufferWidth + x;
        for (i = 0; i < w; i++) {
          pixels24[offset + i] =
              (buf[i * 4 + 2] & 0xFF) << 16 |
              (buf[i * 4 + 1] & 0xFF) << 8 |
              (buf[i * 4] & 0xFF);
        }
        if (rfb.rec != null && !rfb.recordFromBeginning) {
          rfb.rec.write(buf);
        }
      }
    }

    handleUpdatedPixels(x, y, w, h);
    scheduleRepaint(x, y, w, h);
  }

  //
  // Handle a Tight-encoded rectangle.
  //

  void handleTightRect(int x, int y, int w, int h) throws Exception {

    int comp_ctl = rfb.is.readUnsignedByte();
    if (rfb.rec != null) {
      if (rfb.recordFromBeginning ||
          comp_ctl == (RfbProto.TightFill << 4) ||
          comp_ctl == (RfbProto.TightJpeg << 4)) {
        // Send data exactly as received.
        rfb.rec.writeByte(comp_ctl);
      }
      else {
        // Tell the decoder to flush each of the four zlib streams.
        rfb.rec.writeByte(comp_ctl | 0x0F);
      }
    }

    // Flush zlib streams if we are told by the server to do so.
    for (int stream_id = 0; stream_id < 4; stream_id++) {
      if ( (comp_ctl & 1) != 0 && tightInflaters[stream_id] != null) {
        tightInflaters[stream_id] = null;
      }
      comp_ctl >>= 1;
    }

    // Check correctness of subencoding value.
    if (comp_ctl > RfbProto.TightMaxSubencoding) {
      throw new Exception("Incorrect tight subencoding: " + comp_ctl);
    }

    // Handle solid-color rectangles.
    if (comp_ctl == RfbProto.TightFill) {

      if (bytesPixel == 1) {
        int idx = rfb.is.readUnsignedByte();
        memGraphics.setColor(colors[idx]);
        if (rfb.rec != null) {
          rfb.rec.writeByte(idx);
        }
      }
      else {
        byte[] buf = new byte[3];
        rfb.is.readFully(buf);
        if (rfb.rec != null) {
          rfb.rec.write(buf);
        }
        Color bg = new Color(0xFF000000 | (buf[0] & 0xFF) << 16 |
                             (buf[1] & 0xFF) << 8 | (buf[2] & 0xFF));
        memGraphics.setColor(bg);
        buf = null;
      }
      memGraphics.fillRect(x, y, w, h);
      scheduleRepaint(x, y, w, h);
      return;

    }

    if (comp_ctl ==RfbProto.TightJpeg) {

      // Read JPEG data.
      byte[] jpegData = new byte[rfb.readCompactLen()];
      rfb.is.readFully(jpegData);
      if (rfb.rec != null) {
        if (!rfb.recordFromBeginning) {
          rfb.recordCompactLen(jpegData.length);
        }
        rfb.rec.write(jpegData);
      }

      // Create an Image object from the JPEG data.
      Image jpegImage = Toolkit.getDefaultToolkit().createImage(jpegData);

      // Remember the rectangle where the image should be drawn.
      jpegRect = new Rectangle(x, y, w, h);

      // Let the imageUpdate() method do the actual drawing, here just
      // wait until the image is fully loaded and drawn.
      synchronized (jpegRect) {
        Toolkit.getDefaultToolkit().prepareImage(jpegImage, -1, -1, this);
        try {
          // Wait no longer than three seconds.
          jpegRect.wait(3000);
        }
        catch (InterruptedException e) {
          throw new Exception("Interrupted while decoding JPEG image");
        }
      }

      // Done, jpegRect is not needed any more.
      jpegRect = null;
      jpegData = null;
      return;

    }

    // Read filter id and parameters.
    int numColors = 0, rowSize = w;
    byte[] palette8 = new byte[2];
    int[] palette24 = new int[256];
    boolean useGradient = false;
    if ( (comp_ctl &RfbProto.TightExplicitFilter) != 0) {
      int filter_id = rfb.is.readUnsignedByte();
      if (rfb.rec != null) {
        rfb.rec.writeByte(filter_id);
      }
      if (filter_id == RfbProto.TightFilterPalette) {
        numColors = rfb.is.readUnsignedByte() + 1;
        if (rfb.rec != null) {
          rfb.rec.writeByte(numColors - 1);
        }
        if (bytesPixel == 1) {
          if (numColors != 2) {
            throw new Exception("Incorrect tight palette size: " + numColors);
          }
          rfb.is.readFully(palette8);
          if (rfb.rec != null) {
            rfb.rec.write(palette8);
          }
        }
        else {
          byte[] buf = new byte[numColors * 3];
          rfb.is.readFully(buf);
          if (rfb.rec != null) {
            rfb.rec.write(buf);
          }
          for (int i = 0; i < numColors; i++) {
            palette24[i] = ( (buf[i * 3] & 0xFF) << 16 |
                            (buf[i * 3 + 1] & 0xFF) << 8 |
                            (buf[i * 3 + 2] & 0xFF));
          }
          buf = null;
        }
        if (numColors == 2) {
          rowSize = (w + 7) / 8;
        }
      }
      else if (filter_id == RfbProto.TightFilterGradient) {
        useGradient = true;
      }
      else if (filter_id != RfbProto.TightFilterCopy) {
        throw new Exception("Incorrect tight filter id: " + filter_id);
      }
    }
    if (numColors == 0 && bytesPixel == 4) {
      rowSize *= 3;

      // Read, optionally uncompress and decode data.
    }
    int dataSize = h * rowSize;
    if (dataSize < RfbProto.TightMinToCompress) {
      // Data size is small - not compressed with zlib.
      if (numColors != 0) {
        // Indexed colors.
        byte[] indexedData = new byte[dataSize];
        rfb.is.readFully(indexedData);
        if (rfb.rec != null) {
          rfb.rec.write(indexedData);
        }
        if (numColors == 2) {
          // Two colors.
          if (bytesPixel == 1) {
            decodeMonoData(x, y, w, h, indexedData, palette8);
          }
          else {
            decodeMonoData(x, y, w, h, indexedData, palette24);
          }
        }
        else {
          // 3..255 colors (assuming bytesPixel == 4).
          int i = 0;
          for (int dy = y; dy < y + h; dy++) {
            for (int dx = x; dx < x + w; dx++) {
              pixels24[dy * rfb.framebufferWidth + dx] =
                  palette24[indexedData[i++] & 0xFF];
            }
          }
        }
        indexedData = null;
      }
      else if (useGradient) {
        // "Gradient"-processed data
        byte[] buf = new byte[w * h * 3];
        rfb.is.readFully(buf);
        if (rfb.rec != null) {
          rfb.rec.write(buf);
        }
        decodeGradientData(x, y, w, h, buf);
        buf = null;
      }
      else {
        // Raw truecolor data.
        if (bytesPixel == 1) {
          for (int dy = y; dy < y + h; dy++) {
            rfb.is.readFully(pixels8, dy * rfb.framebufferWidth + x, w);
            if (rfb.rec != null) {
              rfb.rec.write(pixels8, dy * rfb.framebufferWidth + x, w);
            }
          }
        }
        else {
          byte[] buf = new byte[w * 3];
          int i, offset;
          for (int dy = y; dy < y + h; dy++) {
            rfb.is.readFully(buf);
            if (rfb.rec != null) {
              rfb.rec.write(buf);
            }
            offset = dy * rfb.framebufferWidth + x;
            for (i = 0; i < w; i++) {
              pixels24[offset + i] =
                  (buf[i * 3] & 0xFF) << 16 |
                  (buf[i * 3 + 1] & 0xFF) << 8 |
                  (buf[i * 3 + 2] & 0xFF);
            }
          }
          buf = null;
        }
      }
    }
    else {
      // Data was compressed with zlib.
      int zlibDataLen = rfb.readCompactLen();
      byte[] zlibData = new byte[zlibDataLen];
      rfb.is.readFully(zlibData);
      if (rfb.rec != null && rfb.recordFromBeginning) {
        rfb.rec.write(zlibData);
      }
      int stream_id = comp_ctl & 0x03;
      if (tightInflaters[stream_id] == null) {
        tightInflaters[stream_id] = new Inflater();
      }
      Inflater myInflater = tightInflaters[stream_id];
      myInflater.setInput(zlibData);
      byte[] buf = new byte[dataSize];
      myInflater.inflate(buf);
      if (rfb.rec != null && !rfb.recordFromBeginning) {
        rfb.recordCompressedData(buf);
      }

      if (numColors != 0) {
        // Indexed colors.
        if (numColors == 2) {
          // Two colors.
          if (bytesPixel == 1) {
            decodeMonoData(x, y, w, h, buf, palette8);
          }
          else {
            decodeMonoData(x, y, w, h, buf, palette24);
          }
        }
        else {
          // More than two colors (assuming bytesPixel == 4).
          int i = 0;
          for (int dy = y; dy < y + h; dy++) {
            for (int dx = x; dx < x + w; dx++) {
              pixels24[dy * rfb.framebufferWidth + dx] =
                  palette24[buf[i++] & 0xFF];
            }
          }
        }
      }
      else if (useGradient) {
        // Compressed "Gradient"-filtered data (assuming bytesPixel == 4).
        decodeGradientData(x, y, w, h, buf);
      }
      else {
        // Compressed truecolor data.
        if (bytesPixel == 1) {
          int destOffset = y * rfb.framebufferWidth + x;
          for (int dy = 0; dy < h; dy++) {
            System.arraycopy(buf, dy * w, pixels8, destOffset, w);
            destOffset += rfb.framebufferWidth;
          }
        }
        else {
          int srcOffset = 0;
          int destOffset, i;
          for (int dy = 0; dy < h; dy++) {
            myInflater.inflate(buf);
            destOffset = (y + dy) * rfb.framebufferWidth + x;
            for (i = 0; i < w; i++) {
              pixels24[destOffset + i] =
                  (buf[srcOffset] & 0xFF) << 16 |
                  (buf[srcOffset + 1] & 0xFF) << 8 |
                  (buf[srcOffset + 2] & 0xFF);
              srcOffset += 3;
            }
          }
        }
      }
      palette8 = null;
      palette24 = null;
      zlibData = null;
      buf = null;
    }

    handleUpdatedPixels(x, y, w, h);
    scheduleRepaint(x, y, w, h);
  }

  //
  // Decode 1bpp-encoded bi-color rectangle (8-bit and 24-bit versions).
  //

  void decodeMonoData(int x, int y, int w, int h, byte[] src, byte[] palette) {

    int dx, dy, n;
    int i = y * rfb.framebufferWidth + x;
    int rowBytes = (w + 7) / 8;
    byte b;

    for (dy = 0; dy < h; dy++) {
      for (dx = 0; dx < w / 8; dx++) {
        b = src[dy * rowBytes + dx];
        for (n = 7; n >= 0; n--) {
          pixels8[i++] = palette[b >> n & 1];
        }
      }
      for (n = 7; n >= 8 - w % 8; n--) {
        pixels8[i++] = palette[src[dy * rowBytes + dx] >> n & 1];
      }
      i += (rfb.framebufferWidth - w);
    }
  }

  void decodeMonoData(int x, int y, int w, int h, byte[] src, int[] palette) {

    int dx, dy, n;
    int i = y * rfb.framebufferWidth + x;
    int rowBytes = (w + 7) / 8;
    byte b;

    for (dy = 0; dy < h; dy++) {
      for (dx = 0; dx < w / 8; dx++) {
        b = src[dy * rowBytes + dx];
        for (n = 7; n >= 0; n--) {
          pixels24[i++] = palette[b >> n & 1];
        }
      }
      for (n = 7; n >= 8 - w % 8; n--) {
        pixels24[i++] = palette[src[dy * rowBytes + dx] >> n & 1];
      }
      i += (rfb.framebufferWidth - w);
    }
  }

  //
  // Decode data processed with the "Gradient" filter.
  //

  void decodeGradientData(int x, int y, int w, int h, byte[] buf) {

    int dx, dy, c;
    byte[] prevRow = new byte[w * 3];
    byte[] thisRow = new byte[w * 3];
    byte[] pix = new byte[3];
    int[] est = new int[3];

    int offset = y * rfb.framebufferWidth + x;

    for (dy = 0; dy < h; dy++) {

      /* First pixel in a row */
      for (c = 0; c < 3; c++) {
        pix[c] = (byte) (prevRow[c] + buf[dy * w * 3 + c]);
        thisRow[c] = pix[c];
      }
      pixels24[offset++] =
          (pix[0] & 0xFF) << 16 | (pix[1] & 0xFF) << 8 | (pix[2] & 0xFF);

      /* Remaining pixels of a row */
      for (dx = 1; dx < w; dx++) {
        for (c = 0; c < 3; c++) {
          est[c] = ( (prevRow[dx * 3 + c] & 0xFF) + (pix[c] & 0xFF) -
                    (prevRow[ (dx - 1) * 3 + c] & 0xFF));
          if (est[c] > 0xFF) {
            est[c] = 0xFF;
          }
          else if (est[c] < 0x00) {
            est[c] = 0x00;
          }
          pix[c] = (byte) (est[c] + buf[ (dy * w + dx) * 3 + c]);
          thisRow[dx * 3 + c] = pix[c];
        }
        pixels24[offset++] =
            (pix[0] & 0xFF) << 16 | (pix[1] & 0xFF) << 8 | (pix[2] & 0xFF);
      }

      System.arraycopy(thisRow, 0, prevRow, 0, w * 3);
      offset += (rfb.framebufferWidth - w);
    }
    prevRow = null;
    thisRow = null;
    pix = null;
    est = null;
  }

  //
  // Display newly updated area of pixels.
  //

  void handleUpdatedPixels(int x, int y, int w, int h) {

    // Draw updated pixels of the off-screen image.
    pixelsSource.newPixels(x, y, w, h);
    memGraphics.setClip(x, y, w, h);
    memGraphics.drawImage(rawPixelsImage, 0, 0, null);
    memGraphics.setClip(0, 0, rfb.framebufferWidth, rfb.framebufferHeight);
  }

  //
  // Tell JVM to repaint specified desktop area.
  //

  void scheduleRepaint(int x, int y, int w, int h) {
    // Request repaint, deferred if necessary.
    repaint(viewer.getDeferScreenUpdates(), x, y, w, h);
  }


  public void keyPressed(KeyEvent evt) {
    processLocalKeyEvent(evt);
  }

  public void keyReleased(KeyEvent evt) {
    processLocalKeyEvent(evt);
  }

  public void keyTyped(KeyEvent evt) {
    evt.consume();
  }
  
//  protected void displayInfo(KeyEvent e, String s){
//      String charString, keyCodeString, modString, tmpString;
//
//      char c = e.getKeyChar();
//      int keyCode = e.getKeyCode();
//      int modifiers = e.getModifiers();
//      
//      if (Character.isISOControl(c)) {
//          charString = "key character = "
//                     + "(an unprintable control character)";
//      } else {
//          charString = "key character = '"
//                     + c + "'";
//      }
//
//      keyCodeString = "key code = " + keyCode
//                      + " ("
//                      + KeyEvent.getKeyText(keyCode)
//                      + ")";
//
//      modString = "modifiers = " + modifiers;
//      tmpString = KeyEvent.getKeyModifiersText(modifiers);
//      if (tmpString.length() > 0) {
//          modString += " (" + tmpString + ")";
//      } else {
//          modString += " (no modifiers)";
//      }
//      
//      log.debug("VncCanvas: " + s + "\n"
//                         + "    " + charString + "\n"
//                         + "    " + keyCodeString + "\n"
//                         + "    " + modString + "\n");
//  }

  public void mousePressed(MouseEvent evt) {
    processLocalMouseEvent(evt, false);
  }

  public void mouseReleased(MouseEvent evt) {
    processLocalMouseEvent(evt, false);
  }

  public void mouseMoved(MouseEvent evt) {
    processLocalMouseEvent(evt, true);
  }

  public void mouseDragged(MouseEvent evt) {
    processLocalMouseEvent(evt, true);
  }

  public void processLocalKeyEvent(KeyEvent evt) {
      if (log.isDebugEnabled()) {
        log.debug("enter :" + "VncCanvas" + ":" + "processLocalKeyEvent");
      }
    if (viewer.getRFB() != null && rfb.inNormalProtocol) {
      if (!inputEnabled) {
        if ( (evt.getKeyChar() == 'r' || evt.getKeyChar() == 'R') &&
            evt.getID() == KeyEvent.KEY_PRESSED) {
          // Request screen update.
          try {
              log.debug("VncCanvas" + ":" + "processLocalKeyEvent"
                    + " --- call writeFramebufferUpdateRequest from refresh");
            rfb.writeFramebufferUpdateRequest(0, 0, rfb.framebufferWidth,
                                              rfb.framebufferHeight, false);
          }
          catch (IOException e) {
            e.printStackTrace();
          }
        }
      }
      else {
        // Input enabled.
        synchronized (rfb) {
          try {
        	if (evt.getKeyCode() != KeyEvent.VK_WINDOWS) {
        		rfb.writeKeyEvent(evt);
        	}
          }
          catch (Exception e) {
            e.printStackTrace();
          }
          /* 
           * No longer processing keyboard keys to pause, button only!
           */
//          if (evt.getKeyChar() == 'p' && evt.getID() == KeyEvent.KEY_PRESSED) {
//              // Keyboard invoked pause event.
//              viewer.getStatusBar().togglePauseMode();
//          }
          rfb.notify();
        }
      }
    }
    // Don't ever pass keyboard events to AWT for default processing.
    // Otherwise, pressing Tab would switch focus to ButtonPanel etc.
    evt.consume();
  }

  public void processLocalMouseEvent(MouseEvent evt, boolean moved) {
    if (viewer.getRFB() != null && rfb.inNormalProtocol) {
      if (moved) {
        softCursorMove(evt.getX(), evt.getY());
      }
      synchronized (rfb) {
        try {
          rfb.writePointerEvent(evt);
        }
        catch (Exception e) {
          e.printStackTrace();
        }
        rfb.notify();
      }
    }
  }

  //
  // Ignored events.
  //

  public void mouseClicked(MouseEvent evt) {
    requestFocus();
  }

  public void mouseEntered(MouseEvent evt) {}

  public void mouseExited(MouseEvent evt) {}

  //////////////////////////////////////////////////////////////////
  //
  // Handle cursor shape updates (XCursor and RichCursor encodings).
  //

  boolean showSoftCursor = false;

  int[] softCursorPixels;
  MemoryImageSource softCursorSource;
  Image softCursor;

  int cursorX = 0, cursorY = 0;
  int cursorWidth, cursorHeight;
  int hotX, hotY;

  //
  // Handle cursor shape update (XCursor and RichCursor encodings).
  //

  synchronized void
      handleCursorShapeUpdate(int encodingType,
                              int xhot, int yhot, int width, int height) throws
      IOException {
      if (log.isDebugEnabled()) {
        log.debug("enter :" + "VncCanvas" + ":" + "handleCursorShapeUpdate");
    }
      
      log.debug("VncCanvas" + ":" + "handleCursorShapeUpdate"
              + " width: " + width + " height: " + height);
      
    int bytesPerRow = (width + 7) / 8;
    int bytesMaskData = bytesPerRow * height;
    
    log.debug("VncCanvas" + ":" + "handleCursorShapeUpdate"
            + " bytesPerRow: " + bytesPerRow + " -- " + " bytesMaskData: " + bytesMaskData + " bytesPixel: " + bytesPixel);

    
    softCursorFree();

    if (width * height == 0) {
      return;
    }

    // Ignore cursor shape data if requested by user.

    if (viewer.getOptions().isIgnoreCursorUpdates()) {
      if (encodingType == RfbProto.EncodingXCursor) {
        rfb.is.skipBytes(6 + bytesMaskData * 2);
      }
      else {
        // rfb.EncodingRichCursor
        rfb.is.skipBytes(width * height + bytesMaskData);
      }
      return;
    }

    // Decode cursor pixel data.

    softCursorPixels = new int[width * height];

    if (encodingType == RfbProto.EncodingXCursor) {

      // Read foreground and background colors of the cursor.
      byte[] rgb = new byte[6];
      rfb.is.readFully(rgb);
      int[] colors = {
          (0xFF000000 | (rgb[3] & 0xFF) << 16 |
           (rgb[4] & 0xFF) << 8 | (rgb[5] & 0xFF)),
          (0xFF000000 | (rgb[0] & 0xFF) << 16 |
           (rgb[1] & 0xFF) << 8 | (rgb[2] & 0xFF))};

      // Read pixel and mask data.
      byte[] pixBuf = new byte[bytesMaskData];
      rfb.is.readFully(pixBuf);
      byte[] maskBuf = new byte[bytesMaskData];
      rfb.is.readFully(maskBuf);

      // Decode pixel data into softCursorPixels[].
      byte pixByte, maskByte;
      int x, y, n, result;
      int i = 0;
      for (y = 0; y < height; y++) {
        for (x = 0; x < width / 8; x++) {
          pixByte = pixBuf[y * bytesPerRow + x];
          maskByte = maskBuf[y * bytesPerRow + x];
          for (n = 7; n >= 0; n--) {
            if ( (maskByte >> n & 1) != 0) {
              result = colors[pixByte >> n & 1];
            }
            else {
              result = 0; // Transparent pixel
            }
            softCursorPixels[i++] = result;
          }
        }
        for (n = 7; n >= 8 - width % 8; n--) {
          if ( (maskBuf[y * bytesPerRow + x] >> n & 1) != 0) {
            result = colors[pixBuf[y * bytesPerRow + x] >> n & 1];
          }
          else {
            result = 0; // Transparent pixel
          }
          softCursorPixels[i++] = result;
        }
      }
      pixBuf = null;
      maskBuf = null;
      colors = null;
    }
    else {
      // encodingType == rfb.EncodingRichCursor

      // Read pixel and mask data.
//      log.debug("VncCanvas" + ":" + "handleCursorShapeUpdate"
//            + " " + "encodingType == rfb.EncodingRichCursor");
//      log.debug("VncCanvas" + ":" + "handleCursorShapeUpdate"
//            + " pixBufSize: " + (width * height * bytesPixel) + " -- "
//            + " maskBufSize: " + (bytesMaskData) + " -- "
//            + " skipBytesSize: " + (width * height + bytesMaskData));

      byte[] pixBuf = new byte[width * height * bytesPixel];
      rfb.is.readFully(pixBuf);
      byte[] maskBuf = new byte[bytesMaskData];
      rfb.is.readFully(maskBuf);
      
      // Debug      
//      log.debug("VncCanvas" + ":" + "handleCursorShapeUpdate"
//            + " firstPixByte: " + pixBuf[0] + " " 
//            + " lastPixByte: " + pixBuf[(width * height * bytesPixel) - 1]);
//      
//      log.debug("VncCanvas" + ":" + "handleCursorShapeUpdate"
//              + " firstMaskByte: " + maskBuf[0] + " " 
//              + " lastMaskByte: " + maskBuf[bytesMaskData - 1]);


      // Decode pixel data into softCursorPixels[].
      byte maskByte;
      int x, y, n, result;
      int i = 0;
      for (y = 0; y < height; y++) {
        for (x = 0; x < width / 8; x++) {
          maskByte = maskBuf[y * bytesPerRow + x];
          for (n = 7; n >= 0; n--) {
            if ( (maskByte >> n & 1) != 0) {
              if (bytesPixel == 1) {
                result = cm8.getRGB(pixBuf[i]);
              }
              else {
                result = 0xFF000000 |
                    (pixBuf[i * 4 + 1] & 0xFF) << 16 |
                    (pixBuf[i * 4 + 2] & 0xFF) << 8 |
                    (pixBuf[i * 4 + 3] & 0xFF);
              }
            }
            else {
              result = 0; // Transparent pixel
            }
            softCursorPixels[i++] = result;
          }
        }
        for (n = 7; n >= 8 - width % 8; n--) {
          if ( (maskBuf[y * bytesPerRow + x] >> n & 1) != 0) {
            if (bytesPixel == 1) {
              result = cm8.getRGB(pixBuf[i]);
            }
            else {
              result = 0xFF000000 |
                  (pixBuf[i * 4 + 1] & 0xFF) << 16 |
                  (pixBuf[i * 4 + 2] & 0xFF) << 8 |
                  (pixBuf[i * 4 + 3] & 0xFF);
            }
          }
          else {
            result = 0; // Transparent pixel
          }
          softCursorPixels[i++] = result;
        }
      }
      pixBuf = null;
      maskBuf = null;
    }

    // Draw the cursor on an off-screen image.

    softCursorSource =
        new MemoryImageSource(width, height, softCursorPixels, 0, width);
    softCursor = createImage(softCursorSource);

    // Set remaining data associated with cursor.

    cursorWidth = width;
    cursorHeight = height;
    hotX = xhot;
    hotY = yhot;

    showSoftCursor = true;

    // Show the cursor.

    repaint(viewer.getDeferCursorUpdates(),
            cursorX - hotX, cursorY - hotY, cursorWidth, cursorHeight);
  }

  //
  // softCursorMove(). Moves soft cursor into a particular location.
  //

  synchronized void softCursorMove(int x, int y) {
    if (showSoftCursor) {
      repaint(viewer.getDeferCursorUpdates(),
              cursorX - hotX, cursorY - hotY, cursorWidth, cursorHeight);
      repaint(viewer.getDeferCursorUpdates(),
              x - hotX, y - hotY, cursorWidth, cursorHeight);
    }

    cursorX = x;
    cursorY = y;
  }

  //
  // softCursorFree(). Remove soft cursor, dispose resources.
  //

  synchronized void softCursorFree() {
      if (log.isDebugEnabled()) {
        log.debug("enter :" + "VncCanvas" + ":" + "softCursorFree");
    }
    if (showSoftCursor) {
        log.debug("VncCanvas" + ":" + "softCursorFree"
                + " showSoftCursor: " + showSoftCursor);
      showSoftCursor = false;
      softCursor = null;
      softCursorSource = null;
      softCursorPixels = null;

      repaint(viewer.getDeferCursorUpdates(),
              cursorX - hotX, cursorY - hotY, cursorWidth, cursorHeight);
    }
  }
}